from __future__ import annotations

import os
import urllib.request
from collections.abc import Sequence
from dataclasses import fields
from pathlib import Path

from omlmd.listener import Event, Listener, PushEvent
from omlmd.model_metadata import ModelMetadata
from omlmd.provider import OMLMDRegistry


def download_file(uri: str):
    file_name = os.path.basename(uri)
    urllib.request.urlretrieve(uri, file_name)
    return file_name


class Helper:
    _listeners: list[Listener] = []

    def __init__(self, registry: OMLMDRegistry | None = None):
        if registry is None:
            self._registry = OMLMDRegistry(
                insecure=True
            )  # TODO: this is a bit limiting when used from CLI, to be refactored
        else:
            self._registry = registry

    @property
    def registry(self):
        return self._registry

    def push(
        self,
        target: str,
        path: Path | str,
        name: str | None = None,
        description: str | None = None,
        author: str | None = None,
        model_format_name: str | None = None,
        model_format_version: str | None = None,
        **kwargs,
    ):
        dataclass_fields = {
            f.name for f in fields(ModelMetadata)
        }  # avoid anything specified in kwargs which would collide
        custom_properties = {
            k: v for k, v in kwargs.items() if k not in dataclass_fields
        }
        model_metadata = ModelMetadata(
            name=name,
            description=description,
            author=author,
            customProperties=custom_properties,
            model_format_name=model_format_name,
            model_format_version=model_format_version,
        )
        if isinstance(path, str):
            path = Path(path)

        json_meta = path.parent / "model_metadata.omlmd.json"
        yaml_meta = path.parent / "model_metadata.omlmd.yaml"
        if (p := json_meta).exists() or (p := yaml_meta).exists():
            raise RuntimeError(
                f"File '{p}' already exists. Aborting TODO: demonstrator."
            )
        json_meta.write_text(model_metadata.to_json())
        yaml_meta.write_text(model_metadata.to_yaml())

        manifest_cfg = f"{json_meta}:application/x-config"
        files = [
            f"{path}:application/x-mlmodel",
            manifest_cfg,
            f"{yaml_meta}:application/x-config",
        ]
        try:
            # print(target, files, model_metadata.to_annotations_dict())
            result = self._registry.push(
                target=target,
                files=files,
                manifest_annotations=model_metadata.to_annotations_dict(),
                manifest_config=manifest_cfg,
            )
            self.notify_listeners(PushEvent(target, model_metadata))
            return result
        finally:
            json_meta.unlink()
            yaml_meta.unlink()

    def pull(
        self, target: str, outdir: Path | str, media_types: Sequence[str] | None = None
    ):
        self._registry.download_layers(target, outdir, media_types)

    def get_config(self, target: str) -> str:
        return f'{{"reference":"{target}", "config": {self._registry.get_config(target)} }}'  # this assumes OCI Manifest.Config later is JSON (per std spec)

    def crawl(self, targets: Sequence[str]) -> str:
        configs = map(self.get_config, targets)
        joined = "[" + ", ".join(configs) + "]"
        return joined

    def add_listener(self, listener: Listener) -> None:
        self._listeners.append(listener)

    def remove_listener(self, listener: Listener) -> None:
        self._listeners.remove(listener)

    def notify_listeners(self, event: Event) -> None:
        for listener in self._listeners:
            listener.update(self, event)



"""Command line interface for OMLMD."""
from __future__ import annotations
from pathlib import Path

import click
import cloup
import logging

from omlmd.helpers import Helper
from omlmd.model_metadata import deserialize_mdfile
from omlmd.provider import OMLMDRegistry


logger = logging.getLogger(__name__)


plain_http = click.option(
    "--plain-http",
    help="Allow insecure connections to registry without SSL check",
    is_flag=True,
    default=False,
    show_default=True,
)


def get_OMLMDRegistry(plain_http: bool) -> OMLMDRegistry:
    return OMLMDRegistry(insecure=plain_http)


@cloup.group()
def cli():
    logging.basicConfig(level=logging.INFO)


@cli.command()
@plain_http
@click.argument("target", required=True)
@click.option(
    "-o",
    "--output",
    default=Path.cwd(),
    show_default=True,
    type=click.Path(path_type=Path, resolve_path=True),
)
@click.option("--media-types", "-m", multiple=True, default=[])
def pull(plain_http: bool, target: str, output: Path, media_types: tuple[str]):
    """Pulls an OCI Artifact containing ML model and metadata, filtering if necessary."""
    Helper(get_OMLMDRegistry(plain_http)).pull(target, output, media_types)


@cli.group()
def get():
    pass


@get.command()
@plain_http
@click.argument("target", required=True)
def config(plain_http: bool, target: str):
    """Outputs configuration of the given OCI Artifact for ML model and metadata."""
    click.echo(Helper(get_OMLMDRegistry(plain_http)).get_config(target))


@cli.command()
@plain_http
@click.argument("targets", required=True, nargs=-1)
def crawl(plain_http: bool, targets: tuple[str]):
    """Crawls configuration for the given list of OCI Artifact for ML model and metadata."""
    click.echo(Helper(get_OMLMDRegistry(plain_http)).crawl(targets))


@cli.command()
@plain_http
@click.argument("target", required=True)
@click.argument(
    "path",
    required=True,
    type=click.Path(path_type=Path, exists=True, resolve_path=True),
)
@cloup.option_group(
    "Metadata options",
    cloup.option(
        "-m",
        "--metadata",
        type=click.Path(path_type=Path, exists=True, resolve_path=True),
        help="Metadata file in JSON or YAML format"
    ),
    cloup.option('--empty-metadata', help='Push with empty metadata', is_flag=True),
    constraint=cloup.constraints.require_one,
)
def push(plain_http: bool, target: str, path: Path, metadata: Path | None, empty_metadata: bool):
    """Pushes an OCI Artifact containing ML model and metadata, supplying metadata from file as necessary"""

    if empty_metadata:
        logger.warning(f"Pushing to {target} with empty metadata.")
    md = deserialize_mdfile(metadata) if metadata else {}
    click.echo(Helper(get_OMLMDRegistry(plain_http)).push(target, path, **md))



from __future__ import annotations

import logging
import os
import tempfile

import oras.defaults
import oras.oci
import oras.provider
import oras.schemas
import oras.utils
from oras.decorator import ensure_container
from oras.provider import container_type

logger = logging.getLogger(__name__)


class OMLMDRegistry(oras.provider.Registry):
    @ensure_container
    def download_layers(self, package, download_dir, media_types):
        """
        Given a manifest of layers, retrieve a layer based on desired media type
        """
        # If you intend to call this function again, you might cache this response
        # for the package of interest.
        manifest = self.get_manifest(package)

        paths = []

        for layer in manifest.get("layers", []):
            if (
                media_types is None
                or len(media_types) == 0
                or layer["mediaType"] in media_types
            ):
                artifact = layer["annotations"]["org.opencontainers.image.title"]
                outfile = oras.utils.sanitize_path(
                    download_dir, os.path.join(download_dir, artifact)
                )
                path = self.download_blob(package, layer["digest"], outfile)
                paths.append(path)

        return paths

    @ensure_container
    def get_config(self, package) -> str:
        """
        Given a manifest of layers, retrieve a layer based on desired media type
        """
        # If you intend to call this function again, you might cache this response
        # for the package of interest.
        manifest = self.get_manifest(package)

        manifest_config = manifest.get("config", {})

        for layer in manifest.get("layers", []):
            if layer["digest"] == manifest_config["digest"]:
                temp_dir = tempfile.mkdtemp()
                try:
                    with tempfile.NamedTemporaryFile(
                        dir=temp_dir, delete=False
                    ) as temp_file:
                        self.download_blob(package, layer["digest"], temp_file.name)
                    with open(temp_file.name, "r") as temp_file_read:
                        file_content = temp_file_read.read()
                        return file_content
                finally:
                    if os.path.exists(temp_dir):
                        for root, dirs, files in os.walk(temp_dir, topdown=False):
                            for file in files:
                                os.remove(os.path.join(root, file))
                            for dir in dirs:
                                os.rmdir(os.path.join(root, dir))
                        os.rmdir(temp_dir)
                        # print("Temporary directory and its contents have been removed.")
        raise RuntimeError("Unable to locate config layer")

    @ensure_container
    def get_manifest_response(
        self,
        container: container_type,
        allowed_media_type: list | None = None,
        refresh_headers: bool = True,
    ) -> dict:
        """
        like get_manifest but return response,
        temporary until https://github.com/oras-project/oras-py/pull/146 in a release.
        """
        if not allowed_media_type:
            allowed_media_type = [oras.defaults.default_manifest_media_type]
        headers = {"Accept": ";".join(allowed_media_type)}

        if not refresh_headers:
            headers.update(self.headers)

        get_manifest = f"{self.prefix}://{container.manifest_url()}"  # type: ignore
        response = self.do_request(get_manifest, "GET", headers=headers)
        self._check_200_response(response)
        return response



from __future__ import annotations

import json
from dataclasses import asdict, dataclass, field, fields
from typing import Any

import yaml


@dataclass
class ModelMetadata:
    name: str | None = None
    description: str | None = None
    author: str | None = None
    customProperties: dict[str, Any] | None = field(default_factory=dict)
    uri: str | None = None
    model_format_name: str | None = None
    model_format_version: str | None = None

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), indent=4)

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)

    def to_annotations_dict(self) -> dict[str, str]:
        as_dict = self.to_dict()
        result = {}
        # TODO: likely key in annotation require FQDN
        for k, v in as_dict.items():
            if isinstance(v, str):
                result[k] = v
            elif v is None:
                continue
            else:
                result[f"{k}+json"] = json.dumps(
                    v
                )  # post-fix "+json" for OCI annotation which is a str representing a json
        return result

    @staticmethod
    def from_json(json_str: str) -> "ModelMetadata":
        data = json.loads(json_str)
        return ModelMetadata(**data)

    def to_yaml(self) -> str:
        return yaml.dump(self.to_dict(), default_flow_style=False)

    @staticmethod
    def from_yaml(yaml_str: str) -> "ModelMetadata":
        data = yaml.safe_load(yaml_str)
        return ModelMetadata(**data)

    @staticmethod
    def from_dict(data: dict[str, Any]) -> "ModelMetadata":
        known_keys = {f.name for f in fields(ModelMetadata)}
        known_properties = {key: data.get(key) for key in known_keys if key in data}
        custom_properties = {
            key: value for key, value in data.items() if key not in known_keys
        }

        return ModelMetadata(**known_properties, customProperties=custom_properties)


def deserialize_mdfile(file):
    with open(file, "r") as file:
        content = file.read()

    try:
        return json.loads(content)
    except json.JSONDecodeError:
        pass

    try:
        return yaml.safe_load(content)
    except yaml.YAMLError:
        pass

    raise ValueError(
        f"The file at {file} is neither a valid JSON nor a valid YAML file."
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from omlmd.model_metadata import ModelMetadata


class Listener(ABC):
    """
    TODO: not yet settled for multi-method or current single update method.
    """

    @abstractmethod
    def update(self, source: Any, event: Event) -> None:
        """
        Receive update event.
        """
        pass


class Event:
    pass


class PushEvent(Event):
    def __init__(self, target: str, metadata: ModelMetadata):
        # TODO: cannot just receive yet the push sha, waiting for: https://github.com/oras-project/oras-py/pull/146 in a release.
        self.target = target
        self.metadata = metadata
